
from .base import BaseCommandEmitter, BaseEventEmitter
from .kafka import KafkaCommandEmitter, KafkaEventEmitter

__all__ = [
    'BaseCommandEmitter',
    'BaseEventEmitter',
    'KafkaCommandEmitter',
    'KafkaEventEmitter'
]



from aiokafka import AIOKafkaConsumer
import logging
import sys
from typing import Any, Dict, Type, Callable, List
import asyncio
from ..serializer.avro import AvroEventSerializer
from ..event.base import BaseEvent
from ..service.command import BaseCommand
from .base import BaseEventConsumer, BaseCommandConsumer
from ..runtime import runtime_context


__all__ = [
    'KafkaEventConsumer',
    'KafkaCommandConsumer'
]


class KafkaEventConsumer(BaseEventConsumer):

    def __init__(self, settings: object, event_topics: List[str] = None, event_group: str = None, position: str = 'earliest'):
        self.logger = logging.getLogger(__name__)
        self.consumer = KafkaConsumer(settings=settings)

        if event_topics:
            self.event_topics = event_topics
        else:
            self.event_topics = settings.EVENTY_EVENT_CONSUMER_SOURCES.split(
                ',')

        self.event_group = event_group
        self.position = position
        self.handlers = dict()

    async def handle_event(self, event: BaseEvent, corr_id: str) -> None:

        handler = self.handlers.get(type(event))
        if handler:
            await handler(event=event, corr_id=corr_id)

    async def start(self):
        await self.consumer.start(
            event_topics=self.event_topics, event_group=self.event_group, event_handler=self.handle_event, position=self.position)

    def register_event_handler(self, event_handler: Callable[[BaseEvent, str], None], event_class: Type[BaseEvent]):
        self.logger.debug(
            f"Registering event handler {event_handler} for event {event_class}")
        self.handlers[event_class] = event_handler


class KafkaCommandConsumer(BaseCommandConsumer):

    def __init__(self, settings: object, event_topic: str = None, event_group: str = None, position: str = 'latest'):
        self.logger = logging.getLogger(__name__)
        self.consumer = KafkaConsumer(settings=settings)

        if event_topic:
            self.event_topic = event_topic
        else:
            self.event_topic = settings.KAFKA_COMMAND_TOPIC

        if event_group:
            self.event_group = event_group
        else:
            self.event_group = settings.KAFKA_COMMAND_GROUP

        self.position = position
        self.command_handlers = dict()

    async def handle_event(self, event: BaseEvent, corr_id: str):
        command_handler = self.command_handlers.get(type(event))
        if command_handler:
            await command_handler(command=event, corr_id=corr_id)

    async def start(self):
        await self.consumer.start(
            event_topics=[self.event_topic], event_group=self.event_group, event_handler=self.handle_event, position=self.position)

    def register_command_handler(self, command_handler: Callable[[BaseCommand, str], None], command_class: Type[BaseCommand]):
        self.logger.debug(
            f"Registering command handler {command_handler} for command {command_class}")
        self.command_handlers[command_class] = command_handler


class KafkaConsumer:

    def __init__(self, settings: Any):
        self.logger = logging.getLogger(__name__)
        self.consumer = None
        self.serializer = runtime_context.serializer
        self.settings = settings

        if self.settings.KAFKA_BOOTSTRAP_SERVER is None:
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')

    async def start(self, event_topics: List[str], event_group: str, event_handler: Callable[[BaseEvent, str], None], position: str):

        consumer_args: Dict[str, Any]
        consumer_args = {
            'loop': asyncio.get_event_loop(),
            'bootstrap_servers': [self.settings.KAFKA_BOOTSTRAP_SERVER],
            'enable_auto_commit': False,
            'group_id': event_group,
            'value_deserializer': self.serializer.decode,
            'auto_offset_reset': position
        }
        if self.settings.KAFKA_USERNAME != '' and self.settings.KAFKA_PASSWORD != '':
            consumer_args.update({
                'sasl_mechanism': 'PLAIN',
                'sasl_plain_username': self.settings.KAFKA_USERNAME,
                'sasl_plain_password': self.settings.KAFKA_PASSWORD,
            })

        try:
            self.logger.info(
                f'Initialize kafka consumer on topic {event_topics}')
            self.consumer = AIOKafkaConsumer(
                *event_topics, **consumer_args)
        except Exception as e:
            self.logger.error(
                f"Unable to connect to the Kafka broker {self.settings.KAFKA_BOOTSTRAP_SERVER} : {e}")
            raise e

        self.logger.info(
            f'Starting kafka consumer on topic {event_topics}')
        try:
            await self.consumer.start()
        except Exception as e:
            self.logger.error(
                f'An error occurred while starting kafka consumer : {e}')
            sys.exit(1)

        try:
            async for msg in self.consumer:
                event = msg.value

                corr_id = event.data['correlation_id']

                await event_handler(event=event, corr_id=corr_id)

                if event_group is not None:
                    await self.consumer.commit()
        except Exception as e:
            self.logger.error(
                f'An error occurred while handling received message : {e}')
            raise e
        finally:
            await self.consumer.stop()



from .base import BaseEventConsumer, BaseCommandConsumer
from .kafka import KafkaEventConsumer, KafkaCommandConsumer


__all__ = [
    'BaseEventConsumer',
    'BaseCommandConsumer',
    'KafkaEventConsumer',
    'KafkaCommandConsumer'
]



from typing import Type, Callable
from ..service.command import BaseCommand
from ..event.base import BaseEvent

__all__ = [
    'BaseEventConsumer',
    'BaseCommandConsumer'
]


class BaseEventConsumer:
    async def start(self):
        raise NotImplementedError

    def register_event_handler(self, event_handler: Callable[[BaseEvent, str], None], event_class: Type[BaseEvent]):
        raise NotImplementedError


class BaseCommandConsumer:

    async def start(self):
        raise NotImplementedError

    def register_command_handler(self, command_handler: Callable[[BaseCommand, str], None], command_class: Type[BaseCommand]):
        raise NotImplementedError



from typing import Any
import logging
from .utils import load_class


logger = logging.getLogger(__name__)


class RuntimeContext:
    def __init__(self):
        self.serializer_class = None
        self.serializer = None
        self.serializer_event_classes = []

        self.command_consumer_class = None
        self.command_consumer = None
        self.command_handlers = []

        self.event_emitter_class = None
        self.event_emitter = None

        self.event_consumer_class = None
        self.event_consumer = None
        self.event_handlers = []

    def init_from_settings(self, settings: Any):

        if settings.EVENTY_SERIALIZER:
            logger.debug(
                f"Initializing serializer {settings.EVENTY_SERIALIZER}")
            self.serializer_class = load_class(settings.EVENTY_SERIALIZER)
            self.serializer = self.serializer_class(settings=settings)
            if len(self.serializer_event_classes) != 0:
                for event_class in self.serializer_event_classes:
                    logger.debug(
                        f"Registering class {event_class['event_class']} for event {event_class['event_name']}")
                    self.serializer.register_event_class(
                        event_class=event_class['event_class'], event_name=event_class['event_name'])

        if settings.EVENTY_COMMAND_CONSUMMER:
            logger.debug(
                f"Initializing command consumer {settings.EVENTY_COMMAND_CONSUMMER}")
            self.command_consumer_class = load_class(
                settings.EVENTY_COMMAND_CONSUMMER)
            self.command_consumer = self.command_consumer_class(
                settings=settings)
            if len(self.command_handlers) != 0:
                for command_handler in self.command_handlers:
                    self.command_consumer.register_command_handler(
                        command_handler=command_handler['command_handler'], command_class=command_handler['command_class'])

        if settings.EVENTY_EVENT_EMITTER:
            self.event_emitter_class = load_class(
                settings.EVENTY_EVENT_EMITTER)
            self.event_emitter = self.event_emitter_class(settings=settings)

        if settings.EVENTY_EVENT_CONSUMER:
            self.event_consumer_class = load_class(
                settings.EVENTY_EVENT_CONSUMER)
            self.event_consumer = self.event_consumer_class(settings=settings)
            if len(self.event_handlers) != 0:
                for event_handler in self.event_handlers:
                    self.event_consumer.register_event_handler(
                        event_handler=event_handler['event_handler'], event_class=event_handler['event_class'])

    def register_event_class(self, event_class, event_name):
        if self.serializer:
            self.serializer.register_event_class(
                event_class=event_class, event_name=event_name)
        else:
            self.serializer_event_classes.append(
                {'event_class': event_class, 'event_name': event_name})

    def register_command_handler(self, command_handler, command_class):
        if self.command_consumer:
            self.command_consumer.register_command_handler(
                command_handler=command_handler, command_class=command_class)
        else:
            self.command_handlers.append(
                {'command_handler': command_handler, 'command_class': command_class})

    def register_event_handler(self, event_handler, event_class):
        if self.event_consumer:
            self.event_consumer.register_event_handler(
                event_handler=event_handler, event_class=event_class)
        else:
            self.event_handlers.append(
                {'event_handler': event_handler, 'event_class': event_class})


runtime_context = RuntimeContext()


def event_class(event_name):
    def decorator(event_class):
        runtime_context.register_event_class(event_class, event_name)
        return event_class
    return decorator


def command_class(command_name):
    return event_class(event_name=command_name)


def command_handler(command_class):
    def decorator(command_handler):
        runtime_context.register_command_handler(
            command_handler, command_class)
        return command_handler
    return decorator


def event_handler(event_class):
    def decorator(event_handler):
        runtime_context.register_event_handler(event_handler, event_class)
        return event_handler
    return decorator



# coding: utf-8
# Copyright (c) Qotto, 2018

from ..utils import current_timestamp, gen_correlation_id

from typing import Any, Dict

__all__ = [
    'BaseEvent'
]


class BaseEvent:
    def __init__(self, data: Dict[str, Any]) -> None:
        if data is None:
            data = dict()

        if 'correlation_id' not in data:
            data['correlation_id'] = gen_correlation_id()

        if 'schema_version' not in data:
            data['schema_version'] = '0.0.0'

        if 'event_timestamp' not in data:
            data['event_timestamp'] = current_timestamp()

        self.name = self.__class__.__name__
        self.data = data

    @classmethod
    def from_data(cls, event_name: str, event_data: Dict[str, Any]):
        raise NotImplementedError



# coding: utf-8
# Copyright (c) Qotto, 2018

from ..event.base import BaseEvent
from ..serializer.base import BaseEventSerializer

__all__ = [
    'BaseEventStore',
]


class BaseEventStore:
    def __init__(self, serializer: BaseEventSerializer) -> None:
        self.serializer = serializer

    def read(self) -> BaseEvent:
        raise NotImplementedError

    def write(self, event: BaseEvent) -> None:
        raise NotImplementedError

    def __iter__(self) -> 'BaseEventStore':
        return self

    def __next__(self) -> BaseEvent:
        return self.read()



from aiokafka import AIOKafkaProducer
import logging
from ..event.base import BaseEvent
from ..service.command import BaseCommand
from ..serializer.base import BaseEventSerializer
from typing import Any, Dict
import asyncio
from .base import BaseEventEmitter, BaseCommandEmitter
from ..runtime import runtime_context

__all__ = [
    'KafkaEventEmitter',
    'KafkaCommandEmitter'
]


class KafkaEventEmitter(BaseEventEmitter):

    def __init__(self, settings=object):
        self.producer = KafkaProducer(settings=settings)
        if settings.KAFKA_EVENT_TOPIC:
            self.default_topic = settings.KAFKA_EVENT_TOPIC

    async def send(self, event: BaseEvent, destination: str = None):
        if destination:
            event_topic = destination
        else:
            event_topic = self.default_topic
        await self.producer.send(event=event, event_topic=event_topic)


class KafkaCommandEmitter(BaseCommandEmitter):
    def __init__(self, settings=object):
        self.producer = KafkaProducer(settings=settings)

    async def send(self, command: BaseCommand, destination: str):
        await self.producer.send(event=command, event_topic=destination)


class KafkaProducer:

    def __init__(self, settings: object):
        self.logger = logging.getLogger(__name__)

        self.serializer = runtime_context.serializer
        self.started = False

        self.settings = settings
        if self.settings.KAFKA_BOOTSTRAP_SERVER is None:
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')
        if self.settings.KAFKA_BOOTSTRAP_SERVER is None:
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')
        if self.settings.KAFKA_BOOTSTRAP_SERVER is None:
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')

        self.producer = self.create_producer()

    async def send(self, event: BaseEvent, event_topic: str):
        if not self.started:
            await self.producer.start()
            self.started = True
        await self.producer.send_and_wait(event_topic, event)

    def create_producer(self) -> AIOKafkaProducer:
        try:

            producer_args: Dict[str, Any]
            producer_args = {
                'loop': asyncio.get_event_loop(),
                'bootstrap_servers': [self.settings.KAFKA_BOOTSTRAP_SERVER],
                'value_serializer': self.serializer.encode
            }
            if self.settings.KAFKA_USERNAME != '' and self.settings.KAFKA_PASSWORD != '':
                producer_args.update({
                    'sasl_mechanism': 'PLAIN',
                    'sasl_plain_username': self.settings.KAFKA_USERNAME,
                    'sasl_plain_password': self.settings.KAFKA_PASSWORD,
                })

            return AIOKafkaProducer(**producer_args)
        except Exception as e:
            self.logger.error(
                f"Unable to connect to the Kafka broker {self.settings.KAFKA_BOOTSTRAP_SERVER}")
            raise e



# coding: utf-8
# Copyright (c) Qotto, 2018

import logging

from ..event.base import BaseEvent
from ..serializer.base import BaseEventSerializer
from .base import BaseEventStore
from kafka import KafkaProducer
from kafka import KafkaConsumer

from typing import Any, Dict, List

__all__ = [
    'KafkaEventStore',
]

logger = logging.getLogger(__name__)


class KafkaEventStore(BaseEventStore):
    """
    An event store that uses Kafka for persistance.
    """
    def __init__(
        self, serializer: BaseEventSerializer,
        bootstrap_servers: List[str], username: str = None, password: str = None,
    ) -> None:
        super().__init__(serializer)
        self._producer: KafkaProducer = None
        self._consumer: KafkaConsumer = None
        self._default_write_topic: str = None
        self._kafka_args: Dict[str, Any] = dict()
        self._kafka_args.update({
            'bootstrap_servers': bootstrap_servers,
        })
        if username and password:
            self._kafka_args.update({
                'sasl_mechanism': 'PLAIN',
                'sasl_plain_username': username,
                'sasl_plain_password': password,
            })

    def register_producer(self, default_write_topic: str = None):
        """
        Registers a Kafka producer. Required to be able to call the `write` method.
        """
        try:
            self._default_write_topic = default_write_topic
            self._producer = KafkaProducer(acks=1, **self._kafka_args)
        except Exception:
            logger.exception(f"Cannot establish connection with Kafka brokers {self._kafka_args['bootstrap_servers']}")
            raise

    def register_consumer(self, group_id: str, topics: str, **args):
        """
        Registers a Kafka consumer. Required to be able to call the `read` method.
        """
        topics_list = [topic.strip() for topic in topics.split(',')]
        try:
            self._consumer = KafkaConsumer(*topics_list, group_id=group_id, **self._kafka_args, **args)
        except Exception:
            logger.exception(f"Cannot establish connection with Kafka brokers {self._kafka_args['bootstrap_servers']}")
            raise

    def read(self) -> BaseEvent:
        if self._consumer is None:
            raise RuntimeError("Kafka consumer is not configured. Please call register_consumer first.")
        serialized_event = next(self._consumer).value
        decoded_event = self.serializer.decode(serialized_event)
        return decoded_event

    def write(self, event: BaseEvent, key: str = None, topic: str = None):
        if self._producer is None:
            raise RuntimeError("Kakfa producer is not configured. Please call register_producer first.")
        if topic is None:
            topic = self._default_write_topic
        serialized_event = self.serializer.encode(event)
        try:
            self._producer.send(topic, serialized_event, key=key)
            self._producer.flush()
        except Exception:
            logger.exception("Failed to dispatch data to Kafka")
            raise
